@@ -24,6 +24,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
+#include <string.h>
 
 #include <string>
 #include <complex>
@@ -72,6 +73,12 @@ int tokenGettoken(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
 int tokenSkip(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals,
 							List* module);
 int tokenSkipCR(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
+int tokenNullLine(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
+int tokenNextChar(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
+int tokenNextStr(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
+int tokenNotNextStr(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
+int tokenNextSyntax(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
+int tokenNotNextSyntax(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals);
 
 std::string	Tokenbuf = "";
 
@@ -84,7 +91,7 @@ int TokenVar(Context* cx, Node* goalscar)
 		
 		Node* env = Nil->Cons(Nil);
 		SetEnv(env, nvar);
-		((Undef*)nvar)->Set(mka((char*)Tokenbuf.c_str()));
+		((Undef*)nvar)->Set(mka(Tokenbuf.c_str()));
 
 		if (cx->tokenflag) cx->token = cx->token + Tokenbuf;
 		Tokenbuf = "";
@@ -269,6 +276,8 @@ int tokenCmpDirect(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals, N
 	}
 
 
+	long tellsave = ftell(cx->ioin);
+	
 	tokenSkipSpace(cx);
 
 	std::string s;
@@ -286,11 +295,8 @@ int tokenCmpDirect(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals, N
 //printf("tokencmpdirect loop %x %x \n", c, 0xff & s[i]);
 		if (c != (0xff & s[i])) {
 //printf("tokencmpdirect if %x %x \n", c, 0xff & s[i]);
-			ungetc(c, cx->ioin);
-			for (i--; i>=0; i--) {
-//printf("tokencmpdirect ungetc %c \n", Tokenbuf[i]);
-				ungetc(Tokenbuf[i], cx->ioin);
-			}
+			fseek(cx->ioin, tellsave, SEEK_SET);
+
 			Tokenbuf = "";
 //printf("tokencmpdirect return -1 %x %x \n", c, 0xff & s[i]);
 			return -1;
@@ -639,13 +645,15 @@ int tokenSpace(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 		return 0;
 	}
 
-	int	i, n;
+	int	n;
 	int ll = ListLength(goalscar);
 	if (ll != 1) {
 		syserr("usage : <SPACE> \n");
 		return 0;
 	}
 
+	long fpos = ftell(cx->ioin);
+
 	int c = fgetc(cx->ioin);
 	if (isSpace(cx, c)) {
 		Node* env=Nil->Cons(Nil);
@@ -653,8 +661,27 @@ int tokenSpace(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 		return 1;
 	}
 
-	ungetc(c, cx->ioin);
-	return -1;
+	char* spc = CharSpace();
+	int  nspc = strlen(spc);
+	if (spc[0] != c) {
+		fseek(cx->ioin, fpos, SEEK_SET);
+		return -1;
+	}
+	
+	char buf[nspc+1];
+	int i;
+	for (i = 1; i < nspc; i++) {
+		c = fgetc(cx->ioin);
+		if (c == EOF) {
+			fseek(cx->ioin, fpos, SEEK_SET);
+			return -1;
+		} else if (c != spc[i]) {
+			fseek(cx->ioin, fpos, SEEK_SET);
+			return -1;
+		}
+	}
+
+	return 1;
 }
 
 int tokenPunct(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
@@ -1263,12 +1290,16 @@ int tokenRangeSub(Context* cx, Node* goalscar, int flg, List* module)
 		} else {
 			syserr("usage : <NONRANGE char char> \n");
 		}
-		return -1;
+		return 0;
 	}
 
 	((Atom*)arg2)->toString(s);
 	c = s[0];
 	n = CharLen(c);
+	if (n > s.length()) {
+		return -1;
+	}
+	
 	cl1 = c & 0xff;
 	for (i = 1; i < n; i++) {
 		cl1 <<= 8;
@@ -1295,12 +1326,16 @@ int tokenRangeSub(Context* cx, Node* goalscar, int flg, List* module)
 		} else {
 			syserr("usage : <NONRANGE char char> \n");
 		}
-		return -1;
+		return 0;
 	}
 	((Atom*)arg3)->toString(s);
 
 	c = s[0];
 	n = CharLen(c);
+	if (n > s.length()) {
+		return -1;
+	}
+	
 	cl2 = c & 0xff;
 	for (i = 1; i < n; i++) {
 		cl2 <<= 8;
@@ -1315,7 +1350,7 @@ int tokenRangeSub(Context* cx, Node* goalscar, int flg, List* module)
 		} else {
 			syserr("usage : <NONRANGE char char> \n");
 		}
-		return -1;
+		return 0;
 	}	
 
 	c = fgetc(cx->ioin);
@@ -1357,7 +1392,6 @@ int tokenRange(Context* cx, Node* goalscar, Node* &goalscdr,
 		syserr("File is not opened");
 		return 0;
 	}
-
 	return tokenRangeSub(cx, goalscar, 1, module);
 }
 
@@ -1368,7 +1402,6 @@ int tokenNonRange(Context* cx, Node* goalscar, Node* &goalscdr,
 		syserr("File is not opened");
 		return 0;
 	}
-
 	return tokenRangeSub(cx, goalscar, 0, module);
 }
 
@@ -1498,4 +1531,302 @@ int tokenSkipCR(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
 }
 
 
+int tokenNullLine(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
+{
+	int c;
+	
+	if (cx->ioin == stdin) {
+		syserr("File is not opened");
+		return 0;
+	}
+
+	int ll = ListLength(goalscar);
+	if (ll != 1) {
+		syserr("usage : <NULLLINE>\n");
+		return 0;
+	}
+
+	tokenSkipSpace(cx);
+	c = fgetc(cx->ioin);
+	if ((c == EOF) || (c == '\n')) {
+		return 1;
+	} 
+	return -1;
+}
+
+
+int tokenNextChar(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
+{
+	if (cx->ioin == stdin) {
+		syserr("File is not opened");
+		return 0;
+	}
+
+	int	i, n;
+	std::string s;
+	
+	int ll = ListLength(goalscar);
+	if ((ll != 2) && (ll != 1)) {
+		syserr("usage : <NEXTCHAR [VAR]> \n");
+		return 0;
+	}
+
+	long fpo = ftell(cx->ioin);
+
+	int c = fgetc(cx->ioin);
+	if (c == EOF) {
+		ungetc(c, cx->ioin);
+		return -1;
+	}
+	
+	std::string buf;
+
+	buf = (char)c;
+	n = CharLen(c);
+	if (n > 1) {
+		for (i=1; i < n; i++) {
+			c = fgetc(cx->ioin);
+			if (c == EOF) {
+				ungetc(c, cx->ioin);
+
+				fseek(cx->ioin, fpo, 0);
+				return -1;
+			}
+			s = (char)c;
+			buf = buf + s;
+		}
+	}
+
+	fseek(cx->ioin, fpo, 0);
+
+	if (ll == 1) {
+		return 1;
+	}
+	
+	if (goalscar->Cdr()->Car()->kind() != UNDEF) {
+		syserr("usage : <NEXTCHAR [VAR]> \n");
+		return 0;
+	}
+
+	Node* nvar = goalscar->Cdr()->Car();
+
+	Node* env=Nil->Cons(Nil);
+	PushStack(cx, goalscar, Nil, env);
+
+	SetEnv(env, nvar);
+	((Undef*)nvar)->Set(mka(buf.c_str()));
+
+	PushStack(cx, Nil, Nil, env);
+	
+	return 1;
+}
+
+
+
+int tokenNextStr(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
+{
+	if (cx->ioin == stdin) {
+		syserr("File is not opened");
+		return 0;
+	}
+
+	int	n;
+	std::string s;
+		
+	int ll = ListLength(goalscar);
+	if (ll != 2) {
+		syserr("usage : <NEXTSTR STRINGS> \n");
+		return 0;
+	}
+
+	if (goalscar->Cdr()->Car()->Val()->kind() != ATOM) {
+		syserr("usage : <NEXTSTR STRINGS> \n");
+		return 0;
+	}
+
+	std::string str;
+	((Atom*)(goalscar->Cdr()->Car()->Val()))->toString(str);
+
+	std::string savecxtoken = cx->token;
+	std::string saveTokenbuf = Tokenbuf;
+	
+	long fpo = ftell(cx->ioin);
+
+	tokenSkipSpace(cx);
+
+	int i, f = 1;
+	for (i = 0; i < str.length(); i++) {
+		int c = fgetc(cx->ioin);
+		if (c == EOF) {
+			f = -1;
+			break;
+		} else if (c != (str[i] & 0xff)) {
+			f = -1;
+			break;
+		}
+	}
+			
+	fseek(cx->ioin, fpo, 0);
+	
+	cx->token = savecxtoken;
+	Tokenbuf = saveTokenbuf;
+		
+	return f;
+}
+
+int tokenNotNextStr(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
+{
+	if (cx->ioin == stdin) {
+		syserr("File is not opened");
+		return 0;
+	}
+
+	int	n;
+	std::string s;
+		
+	int ll = ListLength(goalscar);
+	if (ll != 2) {
+		syserr("usage : <NOTNEXTSTR STRINGS> \n");
+		return 0;
+	}
+
+	if (goalscar->Cdr()->Car()->Val()->kind() != ATOM) {
+		syserr("usage : <NOTNEXTSTR STRINGS> \n");
+		return 0;
+	}
+
+	std::string str;
+	((Atom*)(goalscar->Cdr()->Car()->Val()))->toString(str);
+
+	std::string savecxtoken = cx->token;
+	std::string saveTokenbuf = Tokenbuf;
+	
+	long fpo = ftell(cx->ioin);
+
+	tokenSkipSpace(cx);
+
+	int i, f = -1;
+	for (i = 0; i < str.length(); i++) {
+		int c = fgetc(cx->ioin);
+		if (c == EOF) {
+			f = 1;
+			break;
+		} else if (c != (str[i] & 0xff)) {
+			f = 1;
+			break;
+		}
+	}
+			
+	fseek(cx->ioin, fpo, 0);
+	
+	cx->token = savecxtoken;
+	Tokenbuf = saveTokenbuf;
+		
+	return f;
+}
+
+
+int tokenNextSyntax(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
+{
+	if (cx->ioin == stdin) {
+		syserr("File is not opened");
+		return 0;
+	}
+
+	int	n;
+	std::string s;
+	
+	int ll = ListLength(goalscar);
+	if (ll < 2) {
+		syserr("usage : <NEXTSYNTAX PRED...> \n");
+		return 0;
+	}
+
+	Node* npred = goalscar->Cdr()->Val();
+
+	std::string savecxtoken = cx->token;
+	std::string saveTokenbuf = Tokenbuf;
+	
+	long fpo = ftell(cx->ioin);
+
+	int rn;
+
+	Context *cx2 = new Context(cx->module);
+	cx2->ioin = cx->ioin;
+	cx2->ioout = cx->ioout;
+	cx2->tokenflag = cx->tokenflag;
+	cx2->token = cx->token;
+
+	cxpush(cx2, npred);
+
+	if ((rn=Unify(cx2, npred, cx2->module))>0) {
+		cx->Merge(cx2);
+	}
+
+	cxpop(cx2);
+
+	delete cx2;
+	cx2 = 0;
+
+	fseek(cx->ioin, fpo, 0);
+
+	cx->token = savecxtoken;
+	Tokenbuf = saveTokenbuf;
+	
+	return rn;
+}
+
+
+int tokenNotNextSyntax(Context* cx, Node* goalscar, Node* &goalscdr, Node* &goals)
+{
+	if (cx->ioin == stdin) {
+		syserr("File is not opened");
+		return 0;
+	}
+
+	int	n;
+	std::string s;
+	
+	int ll = ListLength(goalscar);
+	if (ll < 2) {
+		syserr("usage : <NOTNEXTSYNTAX PRED...> \n");
+		return 0;
+	}
+
+	Node* npred = goalscar->Cdr()->Val();
+
+	std::string savecxtoken = cx->token;
+	std::string saveTokenbuf = Tokenbuf;
+	
+	long fpo = ftell(cx->ioin);
+
+	int rn;
+
+	Context *cx2 = new Context(cx->module);
+	cx2->ioin = cx->ioin;
+	cx2->ioout = cx->ioout;
+	cx2->tokenflag = cx->tokenflag;
+	cx2->token = cx->token;
+
+	cxpush(cx2, npred);
+
+	if ((rn=Unify(cx2, npred, cx2->module))>0) {
+		// cx->Merge(cx2);
+	}
+
+	cxpop(cx2);
+
+	delete cx2;
+	cx2 = 0;
+
+	fseek(cx->ioin, fpo, 0);
+
+	cx->token = savecxtoken;
+	Tokenbuf = saveTokenbuf;
+
+	rn = -rn;	
+	return rn;
+}
+
+
 
